/*---------------------------------------------------------------------------\
| PIC18 CAN Peripheral Software for the PIC18F458
|
| File:                 18F_CAN.C           Mainline code
|
| Global Routines:      CanInit()
|                       CanTxIsr()
|                       CanRxIsr()
|                       SendCanMsg()
|                       GetCanMsg()
|
| Local Routines:       StoreMsgInTXBn()
|                       GetMsgFromRXBn() ????
|
| Discussion:           None
|
| Overview:
|
|  Initialization:
|
|   Call InitializeCan to set everything up. See the enums for more details.
|
|  Transmission:
|
|   Mainline calls SendCanMsg which stuffs a message into the one of three
|   queues, CanTxQueue0, CanTxQueue1 or CanTxQueue2 based on the message's
|   ID. CanTx0Isr, CanTx1ISR, and CanTx2ISR then grab the messages from the
|   respective queues and stuff them into the CAN protocol unit's registers
|   to send them (TXB0, TXB1, TXB2). In other words:
|
|                             CanTx2Queue -> CanTx2ISR -> TXB2
|   Main -> SendCanMessage -> CanTx1Queue -> CanTx1ISR -> TXB1
|                             CanTx0Queue -> CanTx0ISR -> TXB0
|
|   We use three buffers because we have lots of RAM and the TXBn buffers
|   are sent in the order TXB2 to TXB0 - thus TXB0 could have very long
|   send latenacies if TXB2 and TXB1 are full. In other words, we
|   guarantee that high prioirty messages are sent to TXB2 and are always
|   sent first, and that medium prioirty messages are sent as soon as
|   high prioirties are sent, etc.
|
|  Reception:
|
|   serial_rx_isr grabs bytes from the UART Receive register (RXREG) on a
|   Receiver Buffer Full Interrupt and stuffs them into the SerialRxQueue.
|   Mainline routines can then grab bytes from the queue using GetSerialByte.
|   In other words:
|
|   RCREG -> serial_rx_isr -> SerialRxQueue -> GetSerialByte -> Main
|
|    Date    | Notes|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/28/2002 | Created.
\---------------------------------------------------------------------------*/


//------------------------------------------------------------------------
//  #includes and local #defines
//------------------------------------------------------------------------

#include "piccore.h"   // Includes for the project (includes other includes!)


//------------------------------------------------------------------------
//  Global Variables (extern'ed in can.h)
//------------------------------------------------------------------------

CanMessage_t    canTx2Queue[CAN_TX2_QUEUE_SIZE];    // CAN Transmit queues.
uint8_t         canTx2IndexOut;
uint8_t         canTx2IndexIn;

CanMessage_t    canTx1Queue[CAN_TX1_QUEUE_SIZE];
uint8_t         canTx1IndexOut;
uint8_t         canTx1IndexIn;

CanMessage_t    canTx0Queue[CAN_TX0_QUEUE_SIZE];
uint8_t         canTx0IndexOut;
uint8_t         canTx0IndexIn;

CanMessage_t    CanRxQueue[CAN_RX_QUEUE_SIZE];
uint8_t         canRxIndexOut;
uint8_t         canRxIndexIn;

bit             canMessageAvailable;

//------------------------------------------------------------------------
//  Local function declarations
//------------------------------------------------------------------------

static void StoreCanMsgId   (uint8_t* pRegister, CanHeader_t value);
static void GetCanMsgId     (uint8_t* pRegister, CanHeader_t *value);



/*---------------------------------------------------------------------------\
| Function:     InitializeCan
|
| Purpose:      Initialize the CAN peripheral
|
| Discussion:   Pass the bitrate and the CAN masks and filters, and it
|               sets up the chip. Many hidden options which shouldn't
|               affect CAN operations hidden here and in SetCantiming.
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

 void InitializeCan (enum CAN_BITRATE_TYPE canBitRate,
                    CanHeader_t canMaskB0,
                    CanHeader_t canMaskB1,
                    CanHeader_t CanFilterB0F0,
                    CanHeader_t CanFilterB0F1,
                    CanHeader_t CanFilterB1F2,
                    CanHeader_t CanFilterB1F3,
                    CanHeader_t CanFilterB1F4,
                    CanHeader_t CanFilterB1F5 )
 {
    // Request config mode - INDETERMINATE TIME CALL!
    // COULD BE 100s of MS! After a reset, we're already
    // in Config Mode so it should take zero time.

    RequestCanMode(CAN_CONFIG_MODE);

    // Now set the CAN Bit rate.

    SetCanTiming(canBitRate);

    // Configure the receive regist. We're hard coding many options
    // since in our case, we'll never need it different (HA!).
    // In particular: Enable double buffer (RB0 ->RB1), receive
    // only valid CAN messages with a standard 11bit identifier,
    // and throw out any currently unread messages (shouldn't be
    // any after a reset).

    RXB0CON = 0b00100100;
          //    ||||||||
          //    |||||||+--- FILHIT0 =  x
          //    ||||||+---- JTOFF   =  x
          //    |||||+----- RXB0DBEN=  1 -> Enable double buffer to RXB1
          //    ||||+------ RXRTRRO =  x
          //    |||+------- <unimplemented on PIC>
          //    ||+-------- RXM0    = 01 -> Receive only valid 11bit MsgIDs
          //    |+--------- RXM1
          //    +---------- RXFUL   =  0 -> (throw out any unread message)

    RXB1CON = 0b00100000;
          //    ||||||||
          //    |||||||+--- FILHIT0 =  x
          //    ||||||+---- FILHIT1 =  x
          //    |||||+----- FILHIT2 =  x
          //    ||||+------ RXRTRRO =  x
          //    |||+------- <unimplemented on PIC>
          //    ||+-------- RXM0    = 01 -> Receive only valid 11bit MsgIDs
          //    |+--------- RXM1
          //    +---------- RXFUL   =  0 -> (throw out any unread message)

    // Set the receive masks and filters.

    CanSetMask(CAN_RX_MASK_B0, canMaskB0);
    CanSetMask(CAN_RX_MASK_B1, canMaskB1);

    CanSetFilter(CAN_RX_FILTER_B0_F0, CanFilterB0F0);
    CanSetFilter(CAN_RX_FILTER_B0_F1, CanFilterB0F1);
    CanSetFilter(CAN_RX_FILTER_B1_F2, CanFilterB1F2);
    CanSetFilter(CAN_RX_FILTER_B1_F3, CanFilterB1F3);
    CanSetFilter(CAN_RX_FILTER_B1_F4, CanFilterB1F4);
    CanSetFilter(CAN_RX_FILTER_B1_F5, CanFilterB1F5);

    // For the love of god, make the TX line ACTIVE HIGH and not
    // OPEN COLLECTOR (hi = tristate).

    ENDRHI = 1;

    // Enable the CAN receive to CCP1 module link (not used for now,
    // but will be used in the future.

    CANCAP = 1;

    // There's nothing else to configure for the transmit side -
    // that's all done when we send a message.

    // Restore to Normal mode.

    RequestCanMode(CAN_NORMAL_MODE);

    // Set the interrupt priority on the CAN interrupts according to config.h

    #if INTS_CAN_TX2_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanTxb2PrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanTxb2PrioritySetting = LOW_PRIORITY_INT;
    #endif

    #if INTS_CAN_TX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanTxb1PrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanTxb1PrioritySetting = LOW_PRIORITY_INT;
    #endif

    #if INTS_CAN_TX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanTxb0PrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanTxb0PrioritySetting = LOW_PRIORITY_INT;
    #endif

    #if INTS_CAN_RX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanRxb0PrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanRxb0PrioritySetting = LOW_PRIORITY_INT;
    #endif

    #if INTS_CAN_RX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanRxb1PrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanRxb1PrioritySetting = LOW_PRIORITY_INT;
    #endif

    #if INTS_CAN_ERR_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
    CanErrPrioritySetting = HIGH_PRIORITY_INT;
    #else
    CanErrPrioritySetting = LOW_PRIORITY_INT;
    #endif

    // DANGER DANGER DANGER DANGER PATCH PATCH FIX ME
    // AAAAAAAAAUGH.

    EnableCanRx0Ints();
    EnableCanRx1Ints();
}







/*---------------------------------------------------------------------------\
| Function:     SetCanTiming
|
| Purpose:      Set the bit rate and timing registers.
|
| Discussion:   Yikes, CAN bus timing is complicated. Here's were we stuff
|               the complexity in all of its glory.
|               Note that we also choose to use the CAN bus line filter
|               for wakeups, choose the # of times to sample, and the SJW
|               here... all of which are hardware-like things that shouldn't
|               need to be changed ever (hah!).
|
| Expects:      To be in configuration mode (else registers will not change
|               their value).
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 02/09/2003 | Added 40MHz code
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/


void SetCanTiming (enum CAN_BITRATE_TYPE canBitRate)
{
    switch (canBitRate)
    {
        case (CAN_BITRATE_125KBPS):

            // 125kbps per second -> Tbit = 8us
            // ASSUME 20MHz CLOCK.
            // Divide by 10 to give us 16 CAN clocks (Tq's) in 1 bit.
            // Meaning 1Tq = 500ns.
            //
            //  |---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|---|
            //  |   |Prop   | Seg1                      |Seg2                       |
            //  Sync                                    ^Sample = 62.5%
            //
            // which translates to a SEG1 = 7Tq and SEG2 = 6T if the Sync is 1Tq
            // and PROP is 2Tq.

            // The values to set the registers are really weird, so we'll set
            // them explicitly here and not try and hide the weirdness:

#if PROCESSOR_CLOCK_FREQ == 20

            BRGCON1 = 0b00000100;
                  //    ||||||||
                  //    |||||||+--- BRP0 = 000100 = Tq = 2*5/FOSC = 500ns -> 16Tq/bit
                  //    ||||||+---- BRP1
                  //    |||||+----- BRP2
                  //    ||||+------ BRP3
                  //    |||+------- BRP4
                  //    ||+-------- BRP5
                  //    |+--------- SJW = 00 -> SJW = 1Tq
                  //    +---------- SJW1

#elif PROCESSOR_CLOCK_FREQ == 40

            BRGCON1 = 0b00001001;
                  //    ||||||||
                  //    |||||||+--- BRP0 = 001001 = Tq = 2*10/FOSC = 500ns -> 16Tq/bit
                  //    ||||||+---- BRP1
                  //    |||||+----- BRP2
                  //    ||||+------ BRP3
                  //    |||+------- BRP4
                  //    ||+-------- BRP5
                  //    |+--------- SJW = 00 -> SJW = 1Tq
                  //    +---------- SJW1

#else

    #error can.c: HEY! THE CLOCK FREQUENCY MUST BE EITHER 20 OR 40

#endif

            BRGCON2 = 0b11110001;
                  //    ||||||||
                  //    |||||||+--- PRSEG0 = 001    -> PROP = 2Tq
                  //    ||||||+---- PRSEG1
                  //    |||||+----- PRSEG2
                  //    ||||+------ SEG1PH0 = 110   -> SEG1 = 7Tq
                  //    |||+------- SEG1PH1
                  //    ||+-------- SEG1PH2
                  //    |+--------- SAM = 1         -> three samples
                  //    +---------- SEG2PHTS = 1    -> Use SEG2PH bits

            BRGCON3 = 0b10100010;
                  //    ||||||||
                  //    |||||||+--- <Unimplemented on PIC>
                  //    ||||||+---- WAKFIL = 1      -> Use CAN filter on wakeup
                  //    |||||+----- <Unimplemented on PIC>
                  //    ||||+------ <Unimplemented on PIC>
                  //    |||+------- <Unimplemented on PIC>
                  //    ||+-------- SEG2PH0 = 101   -> SEG2 = 6Tq
                  //    |+--------- SEG2PH1
                  //    +---------- SEG2PH2

            break;

        case (CAN_BITRATE_1MBPS):
        default:

            // 1Mb per second -> Tbit = 1us
            // Fosc: Assume 20MHz or 40MHz / 2, which gives us the same Fosc.
            // Divide by 2 to give us 10 CAN clocks (Tq's) in 1 bit.
            // Meaning 1Tq = 100ns.
            //
            // IN LV2's case, the Flight Computer (Intel 82527) does 8Tq/bit:
            //
            //  |--------|--------|--------|--------|-------|--------|--------|-------|
            //  |Tsync   |Prop + Seg1                       |Seg2                     |
            //                                              ^Sample = 62.5%
            //
            // And the closest match we can do it at 10Tq/bit is:
            //
            //  |------|------|------|------|------|------|------|------|------|------|
            //  |Tsync |Prop  |Seg1                       |Seg2                       |
            //                                            ^Sample = 60.0%
            //
            // which translates to a SEG1 = 4Tq and SEG2 = 4T if the PROP and Tsync
            // segments are both 1Tq.

            // The values to set the registers are really weird, so we'll set
            // them explicitly here and not try and hide the weirdness:

#if PROCESSOR_CLOCK_FREQ == 20

            BRGCON1 = 0b00000000;
                  //    ||||||||
                  //    |||||||+--- BRP0 = 000000 = Tq = 2/FOSC = 100ns -> 10Tq/bit
                  //    ||||||+---- BRP1
                  //    |||||+----- BRP2
                  //    ||||+------ BRP3
                  //    |||+------- BRP4
                  //    ||+-------- BRP5
                  //    |+--------- SJW = 00 -> SJW = 1Tq
                  //    +---------- SJW1

#elif PROCESSOR_CLOCK_FREQ == 40

            BRGCON1 = 0b00000001;
                  //    ||||||||
                  //    |||||||+--- BRP0 = 000001 = Tq = 4/FOSC = 100ns -> 10Tq/bit
                  //    ||||||+---- BRP1
                  //    |||||+----- BRP2
                  //    ||||+------ BRP3
                  //    |||+------- BRP4
                  //    ||+-------- BRP5
                  //    |+--------- SJW = 00 -> SJW = 1Tq
                  //    +---------- SJW1

#else

    #error can.c: HEY! THE CLOCK FREQUENCY MUST BE EITHER 20 OR 40!

#endif

            BRGCON2 = 0b10011000;
                  //    ||||||||
                  //    |||||||+--- PRSEG0 = 000    -> PROP = 1Tq
                  //    ||||||+---- PRSEG1
                  //    |||||+----- PRSEG2
                  //    ||||+------ SEG1PH0 = 011   -> SEG1 = 4Tq
                  //    |||+------- SEG1PH1
                  //    ||+-------- SEG1PH2
                  //    |+--------- SAM = 0         -> one sample
                  //    +---------- SEG2PHTS = 1    -> Use SEG2PH bits

            BRGCON3 = 0b01000011;
                  //    ||||||||
                  //    |||||||+--- SEG2PH0 = 011   -> SEG2 = 4Tq
                  //    ||||||+---- SEG2PH1
                  //    |||||+----- SEG2PH2
                  //    ||||+------ <Unimplemented on PIC>
                  //    |||+------- <Unimplemented on PIC>
                  //    ||+-------- <Unimplemented on PIC>
                  //    |+--------- WAKFIL = 1      -> Use CAN filter on wakeup
                  //    +---------- <Unimplemented on PIC>

            if (canBitRate != CAN_BITRATE_1MBPS)
                ThrowWarning(WARN_CAN_BITRATE);
            break;
    }
}





/*---------------------------------------------------------------------------\
| Function:     RequestCanMode
|
| Purpose:      Set the CAN mode bits and hang around until they change.
|
| Discussion:   WARNING! Hanging around until they change could be a really
|               long time... hundreds of ms to even seconds. EVENTUALLY, THIS
|               ROUTINE SHOULD HAVE A TIME OUT.
|
| Expects:      The wait to be ok
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void RequestCanMode (enum CAN_MODES mode)
{
    // Request desired mode.

    CANCON = mode;

    // Wait till desired mode is set.

    while( (  CANSTAT & CAN_MODE_BITS) != mode );
}



/*---------------------------------------------------------------------------\
| Function:     CanSetMask
|
| Purpose:      Set the CAN mask for RB0 or RB1 with the specified mask value
|
| Discussion:   Passes the 16bit mask value for either the RB0 or RB1 mask to
|               the routine StoreCanMsgId which then updates all 11bits of the
|               mask.
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/


void CanSetMask(enum CAN_MASK canMaskNum,CanHeader_t mask )
{
    // Store the 11bit mask in the correct Receive Buffer
    // mask given by canMaskNum. Pass the pointer to the
    // correct register, assuming StoreCanMsgId realizes
    // the upper bits of the pointer are for the SFR bank.

    switch(canMaskNum)
    {

        case CAN_RX_MASK_B0:
        StoreCanMsgId((uint8_t*)&RXM0SIDH, mask);
            break;

        case CAN_RX_MASK_B1:
        StoreCanMsgId((uint8_t*)&RXM1SIDH, mask);
            break;

        default:
            ThrowWarning(WARN_CAN_MASK_NUM);
            break;
    }
}





/*---------------------------------------------------------------------------\
| Function:     CanSetMask
|
| Purpose:      Set the CAN mask for RB0 or RB1 with the specified mask value
|
| Discussion:   Passes the 16bit mask value for either the RB0 or RB1 mask to
|               the routine StoreCanMsgId which then updates all 11bits of the
|               mask.
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void CanSetFilter ( enum CAN_FILTER canFilterNum,CanHeader_t filter)
{
    switch(canFilterNum)
    {

    case CAN_RX_FILTER_B0_F0:
        StoreCanMsgId((uint8_t*)&RXF0SIDH, filter);
        break;

    case CAN_RX_FILTER_B0_F1:
        StoreCanMsgId((uint8_t*)&RXF1SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F2:
        StoreCanMsgId((uint8_t*)&RXF2SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F3:
        StoreCanMsgId((uint8_t*)&RXF3SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F4:
        StoreCanMsgId((uint8_t*)&RXF4SIDH, filter);
        break;

    case CAN_RX_FILTER_B1_F5:
        StoreCanMsgId((uint8_t*)&RXF5SIDH, filter);
        break;

    default:
        ThrowWarning(WARN_CAN_FILTER_NUM);
        break;
    }
}

/*---------------------------------------------------------------------------\
| Function:     StoreCanMsgId
|
| Purpose:      Store an 11bit value into 2 8bit registers
|
| Discussion:   Passes 16bit address of 8bit register to be updates - the
|               routine StoreCanMsgId then updates it and it's higher-in-
|               memory 8bit partner so that all 11bits of the mask are stored.
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

static void StoreCanMsgId(uint8_t* pRegister, CanHeader_t value )
{
    // Take bits 15:8 of the CanHeader_t and write them to the high byte
    // of the SFR register pair which holds a 11bits

    *pRegister =  value.bytes.high;

    // Point to the low byte (always one higher in the SFRs) of the
    // SFR register pair.

    pRegister++;

    // Take bits 7:5 of the CanHeader_t and write them to the low bute
    // of the SFR which also uses 7:5. And we can ignore 4:0 since
    // It's always extended (29bit) ID bits which we don't care about.

    *pRegister = value.bytes.low;

    GetCanMsgId(pRegister, &value);
}



/*---------------------------------------------------------------------------\
| Function:     GetCanMsgId
|
| Purpose:      Get an 11bit value packed in 2 8bit registers
|
| Discussion:   It's 4:00am. I feel like crap. How are you?
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 05/21/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

static void GetCanMsgId(uint8_t* pRegister, CanHeader_t *pValue)
{
    // Get bits 10:3 from the SFR register and store it in the high
    // byte of the CanHeader_t

    pValue->bytes.high = *pRegister;

    // Point to the next register of the pair (less signficant)

    pRegister++;

    // Get bits 2:0 from the LS SFR register and clear the rest of the
    // bits just in case.

    pValue->bytes.low = (*pRegister & 0b11100000);
}



/*---------------------------------------------------------------------------\
| Function:     SendCanMessage
|
| Purpose:      Load a CAN message into one of three CAN message queues.
|
| Discussion:   Chooses which queue to put the message in.
|               Disables CAN TX Buffer <n> empty interrupts for <N> cycles.
|               Sure we could encapsulate the message writes into a seperate
|               function calls but lets not waste cycles for the sake of
|               clearness ;)
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

void SendCanMessage (CanMessage_t* pMessage)
{

    static  uint8_t     i;
    static  uint16_t    id;


    // First, make sure we're not sending bad messages.

    if ( pMessage->header.bits.length > 8)
    {
        ThrowWarning(WARN_CAN_BAD_TX_DLC);
        return;
    }

    id = (pMessage->header.all & 0b1111111111100000); // FIXES WEIRD "0" CONSTANT PROBLEM.

    if ( (id < MIN_MESSAGE_ID) || (id > MAX_MESSAGE_ID) )
    {
        ThrowWarning(WARN_CAN_BAD_TX_ID);
        return;
    }

    // if Message's ID priority  >= high priority (i.e., it's LOWER)
    // then load message into canTx2Queue

    if ( id <= HIGH_PRIORITY_TX_CUTOFF )
    {
        DisableCanTx2Ints();

        // Copy over header info (ID, RTR, DLC)

        canTx2Queue[canTx2IndexIn].header.bytes.low = pMessage->header.bytes.low;
        canTx2Queue[canTx2IndexIn].header.bytes.high = pMessage->header.bytes.high;

        // Copy over all data bytes

        for (i = 0; i < pMessage->header.bits.length; ++i)
            canTx2Queue[canTx2IndexIn].data[i] = pMessage->data[i];

        // Increment the in index

        if ( canTx2IndexIn < (CAN_TX2_QUEUE_SIZE-1) )
            ++canTx2IndexIn;
        else
            canTx2IndexIn = 0;

        // Check for overwrites. If we incremented the indexIn and it now
        // equals the indexOut then we've wrapped around the queue and
        // we have a queue overrun. In this case, don't write to the
        // queue - back off by one and throw a warning.

        if (canTx2IndexIn == canTx2IndexOut)
        {
            if (canTx2IndexIn == 0)
                canTx2IndexIn  = (CAN_TX2_QUEUE_SIZE-1);
            else
                --canTx2IndexIn;

            ThrowWarning(WARN_CAN_TX2_QUEUE_OVERFLOW);
        }

        // Now we've loaded a message into the queue. If there's already a
        // message being transmitted in TXB2, then when it's finished CanTx2Isr will
        // automatically grab the next. But if TXB2 has been idle at all, then
        // there's no way for CanTx2Isr to know that the queue has a message in it.
        // So we check to see if the buffer is idle, and if it is, we force an
        // interrupt by writing directly to the TXB2 interrupt flag.

        if ( CanTx2Full == FALSE )
            ForceCanTx2Interrupt();

        // Whatever the case, enable the CAN TXB2 interrupt.

        EnableCanTx2Ints();
    }

    // else if Message's ID priority >= low prioirty (i.e., it's LOWER)
    // then load message into canTx1Queue

    else if ( id <= NORMAL_PRIORITY_TX_CUTOFF )
    {
        DisableCanTx1Ints();

        // Copy over header info (ID, RTR, DLC)

        canTx1Queue[canTx1IndexIn].header.bytes.low = pMessage->header.bytes.low;
        canTx1Queue[canTx1IndexIn].header.bytes.high = pMessage->header.bytes.high;

        // Copy over all data bytes

        for (i = 0; i < pMessage->header.bits.length; ++i)
            canTx1Queue[canTx1IndexIn].data[i] = pMessage->data[i];

        // Increment the in index

        if ( canTx1IndexIn < (CAN_TX1_QUEUE_SIZE-1) )
            ++canTx1IndexIn;
        else
            canTx1IndexIn = 0;

        // Check for overwrites. If we incremented the indexIn and it now
        // equals the indexOut then we've wrapped around the queue and
        // we have a queue overrun. In this case, don't write to the
        // queue - back off by one and throw a warning.

        if (canTx1IndexIn == canTx1IndexOut)
        {
            if (canTx1IndexIn == 0)
                canTx1IndexIn  = (CAN_TX1_QUEUE_SIZE-1);
            else
                --canTx1IndexIn;

            ThrowWarning(WARN_CAN_TX1_QUEUE_OVERFLOW);
        }

        // Now we've loaded a message into the queue. If there's already a
        // message being transmitted in TXB1, then when it's finished CanTx2Isr will
        // automatically grab the next. But if TXB1 has been idle at all, then
        // there's no way for CanTx1Isr to know that the queue has a message in it.
        // So we check to see if the buffer is idle, and if it is, we force an
        // interrupt by writing directly to the TXB1 interrupt flag.

        if ( CanTx1Full == FALSE )
            ForceCanTx1Interrupt();

        // Whatever the case, enable the CAN TXB2 interrupt.

        EnableCanTx1Ints();
    }

    // else load message into CanTx0Queue (must be a low prioirty message)

    else
    {
        DisableCanTx0Ints();

        // Copy over header info (ID, RTR, DLC)

        canTx0Queue[canTx0IndexIn].header.bytes.low = pMessage->header.bytes.low;
        canTx0Queue[canTx0IndexIn].header.bytes.high = pMessage->header.bytes.high;

        // Copy over all data bytes

        for (i = 0; i < pMessage->header.bits.length; ++i)
            canTx0Queue[canTx0IndexIn].data[i] = pMessage->data[i];

        // Increment the in index

        if ( canTx0IndexIn < (CAN_TX0_QUEUE_SIZE-1) )
            ++canTx0IndexIn;
        else
            canTx0IndexIn = 0;

        // Check for overwrites. If we incremented the indexIn and it now
        // equals the indexOut then we've wrapped around the queue and
        // we have a queue overrun. In this case, don't write to the
        // queue - back off by one and throw a warning.

        if (canTx0IndexIn == canTx0IndexOut)
        {
            if (canTx0IndexIn == 0)
                canTx0IndexIn  = (CAN_TX0_QUEUE_SIZE-1);
            else
                --canTx0IndexIn;

            ThrowWarning(WARN_CAN_TX0_QUEUE_OVERFLOW);
        }

        // Now we've loaded a message into the queue. If there's already a
        // message being transmitted in TXB0, then when it's finished CanTx2Isr will
        // automatically grab the next. But if TXB0 has been idle at all, then
        // there's no way for CanTx2Isr to know that the queue has a message in it.
        // So we check to see if the buffer is idle, and if it is, we force an
        // interrupt by writing directly to the TXB0 interrupt flag.

        if ( CanTx0Full == FALSE )
            ForceCanTx0Interrupt();

        // Whatever the case, enable the CAN TXB0 interrupt.

        EnableCanTx0Ints();
    }
}



/*---------------------------------------------------------------------------\
| Function:     CanTx2Isr
|
| Purpose:      Grab messages from the CanTx2Queue and load them into TXB2.
|               Sets interrupt on exit if the CAN unit is currently idle.
|
| Discussion:
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

#if INTS_CAN_TX2_ISR_USED == INTS_CORE_ISR_ENABLED

void CanTx2Isr (void)
{
    static uint8_t* pRegister;
    static uint8_t  i;


    // Clear the interrupt

    ClearCanTx2Interrupt();

    // If we're here, someone has PROBABLY written to the canTx2Queue and enabled
    // interrupts (See SendCanMessage()). We can assume that
    // canTx2IndexOut points to one past the last message transmitted
    // (e.g., points to the next message to transmit).

    // If the in and out indices are equal, however, there aren't any more messages
    // in the transmit queue so we MUST turn off interrupts otherwise
    // we'll keep coming back here and thus lock up the PIC.

    if (TXB2REQ == 1)
    {
        #if INTS_CAN_TX2_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(WARN_CAN_TXB2_NOT_EMPTY);
        #else
        ThrowWarning_LP(WARN_CAN_TXB2_NOT_EMPTY);
        #endif
        DisableCanTx2Ints();
    }

    else
    {
        if (canTx2IndexOut == canTx2IndexIn)
        {
            #if INTS_CAN_TX2_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
            ThrowWarning_HP(WARN_CAN_TX2_QUEUE_EMPTY);
            #else
            ThrowWarning_LP(WARN_CAN_TX2_QUEUE_EMPTY);
            #endif
            DisableCanTx2Ints();
        }

        else
        {
            // Grab a message from canTx2Queue and stuff it into TXB2. Note that we're
            // not going to bother to setup CANCON to move the TXB2 registers into the
            // access bank - although for future renditions this could be a good idea.

            // Copy over the message ID

            TXB2SIDL = canTx2Queue[canTx2IndexOut].header.bytes.low & 0b11100000;
            TXB2SIDH = canTx2Queue[canTx2IndexOut].header.bytes.high;

            // Copy over the DLC and RTR bits

            TXB2DLC =  canTx2Queue[canTx2IndexOut].header.bytes.low & 0b00001111;
            TXB2RTR = canTx2Queue[canTx2IndexOut].header.bits.request;

            // Copy over all data bytes

            pRegister = &TXB2D0;

            for (i = 0; i < canTx2Queue[canTx2IndexOut].header.bits.length; ++i)
                *pRegister++ = canTx2Queue[canTx2IndexOut].data[i];

            // And send that puppy!

            TXB2REQ = 1;

            // Ok, it's out of our hands now... Go CAN Protocol Unit go!
            // Increment the out index to the CIRCULAR queue.

            if (canTx2IndexOut < (CAN_TX2_QUEUE_SIZE-1))
                ++canTx2IndexOut;
            else
                canTx2IndexOut = 0;

            // We've increased the counter, so check to see if the two pointers are equal.
            // If they are, then there's nothing in canTx2Queue to send to turn off transmit
            // buffer empty interrupts.

            if (canTx2IndexOut == canTx2IndexIn)
                DisableCanTx2Ints();
        }
    }
}

#endif //INTS_CAN_TX2_ISR_USED == INTS_CORE_ISR_ENABLED


/*---------------------------------------------------------------------------\
| Function:     CanTx1Isr
|
| Purpose:      Grab messages from the CanTx1Queue and load them into TXB1.
|               Sets interrupt on exit if the CAN unit is currently idle.
|
| Discussion:
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

#if INTS_CAN_TX1_ISR_USED == INTS_CORE_ISR_ENABLED

void CanTx1Isr (void)
{
    static uint8_t* pRegister;
    static uint8_t  i;


    // Clear the interrupt

    ClearCanTx1Interrupt();

    // If we're here, someone has PROBABLY written to the canTx1Queue and enabled
    // interrupts (See SendCanMessage()). We can assume that
    // canTx1IndexOut points to one past the last message transmitted
    // (e.g., points to the next message to transmit).

    // If the in and out indices are equal, however, there aren't any more messages
    // in the transmit queue so we MUST turn off interrupts otherwise
    // we'll keep coming back here and thus lock up the PIC.

    if (TXB1REQ == 1)
    {
        #if INTS_CAN_TX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(WARN_CAN_TXB1_NOT_EMPTY);
        #else
        ThrowWarning_LP(WARN_CAN_TXB1_NOT_EMPTY);
        #endif
        DisableCanTx1Ints();
    }

    else
    {
        if (canTx1IndexOut == canTx1IndexIn)
        {
            #if INTS_CAN_TX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
            ThrowWarning_HP(WARN_CAN_TX1_QUEUE_EMPTY);
            #else
            ThrowWarning_LP(WARN_CAN_TX1_QUEUE_EMPTY);
            #endif
            DisableCanTx1Ints();
        }

        else
        {
            // Grab a message from canTx1Queue and stuff it into TXB1. Note that we're
            // not going to bother to setup CANCON to move the TXB1 registers into the
            // access bank - although for future renditions this could be a good idea.

            // Copy over the message ID

            TXB1SIDL = canTx1Queue[canTx1IndexOut].header.bytes.low & 0b11100000;
            TXB1SIDH = canTx1Queue[canTx1IndexOut].header.bytes.high;

            // Copy over the DLC and RTR bits

            TXB1DLC =  canTx1Queue[canTx1IndexOut].header.bytes.low & 0b00001111;
            TXB1RTR = canTx1Queue[canTx1IndexOut].header.bits.request;

            // Copy over all data bytes

            pRegister = &TXB1D0;

            for (i = 0; i < canTx1Queue[canTx1IndexOut].header.bits.length; ++i)
                *pRegister++ = canTx1Queue[canTx1IndexOut].data[i];

            // And send that puppy!

            TXB1REQ = 1;

            // Ok, it's out of our hands now... Go CAN Protocol Unit go!
            // Increment the out index to the CIRCULAR queue.

            if (canTx1IndexOut < (CAN_TX1_QUEUE_SIZE-1))
                ++canTx1IndexOut;
            else
                canTx1IndexOut = 0;

            // We've increased the counter, so check to see if the two pointers are equal.
            // If they are, then there's nothing in canTx1Queue to send to turn off transmit
            // buffer empty interrupts.

            if (canTx1IndexOut == canTx1IndexIn)
                DisableCanTx1Ints();
        }
    }
}

#endif //INTS_CAN_TX1_ISR_USED == INTS_CORE_ISR_ENABLED


/*---------------------------------------------------------------------------\
| Function:     CanTx0Isr
|
| Purpose:      Grab messages from the CanTx0Queue and load them into TXB0.
|               Sets interrupt on exit if the CAN unit is currently idle.
|
| Discussion:
|
| Expects:
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/30/2002 | Created if you can call it that.
\---------------------------------------------------------------------------*/

#if INTS_CAN_TX0_ISR_USED == INTS_CORE_ISR_ENABLED

void CanTx0Isr (void)
{
    static uint8_t* pRegister;
    static uint8_t  i;


    // Clear the interrupt

    ClearCanTx0Interrupt();

    // If we're here, someone has PROBABLY written to the canTx0Queue and enabled
    // interrupts (See SendCanMessage()). We can assume that
    // canTx0IndexOut points to one past the last message transmitted
    // (e.g., points to the next message to transmit).

    // If the in and out indices are equal, however, there aren't any more messages
    // in the transmit queue so we MUST turn off interrupts otherwise
    // we'll keep coming back here and thus lock up the PIC.

    if (TXB0REQ == 1)
    {
        #if INTS_CAN_TX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(WARN_CAN_TXB0_NOT_EMPTY);
        #else
        ThrowWarning_LP(WARN_CAN_TXB0_NOT_EMPTY);
        #endif
        DisableCanTx0Ints();
    }

    else
    {
        if (canTx0IndexOut == canTx0IndexIn)
        {
            #if INTS_CAN_TX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
            ThrowWarning_HP(WARN_CAN_TX0_QUEUE_EMPTY);
            #else
            ThrowWarning_LP(WARN_CAN_TX0_QUEUE_EMPTY);
            #endif
            DisableCanTx0Ints();
        }

        else
        {
            // Grab a message from canTx0Queue and stuff it into TXB0. Note that we're
            // not going to bother to setup CANCON to move the TXB0 registers into the
            // access bank - although for future renditions this could be a good idea.

            // Copy over the message ID

            TXB0SIDL = canTx0Queue[canTx0IndexOut].header.bytes.low & 0b11100000;
            TXB0SIDH = canTx0Queue[canTx0IndexOut].header.bytes.high;

            // Copy over the DLC and RTR bits

            TXB0DLC =  canTx0Queue[canTx0IndexOut].header.bytes.low & 0b00001111;
            TXB0RTR = canTx0Queue[canTx0IndexOut].header.bits.request;

            // Copy over all data bytes

            pRegister = &TXB0D0;

            for (i = 0; i < canTx0Queue[canTx0IndexOut].header.bits.length; ++i)
                *pRegister++ = canTx0Queue[canTx0IndexOut].data[i];

            // And send that puppy!

            TXB0REQ = 1;

            // Ok, it's out of our hands now... Go CAN Protocol Unit go!
            // Increment the out index to the CIRCULAR queue.

            if (canTx0IndexOut < (CAN_TX0_QUEUE_SIZE-1))
                ++canTx0IndexOut;
            else
                canTx0IndexOut = 0;

            // We've increased the counter, so check to see if the two pointers are equal.
            // If they are, then there's nothing in canTx0Queue to send to turn off transmit
            // buffer empty interrupts.

            if (canTx0IndexOut == canTx0IndexIn)
                DisableCanTx0Ints();
        }
    }
}

#endif //INTS_CAN_TX0_ISR_USED == INTS_CORE_ISR_ENABLED



/*---------------------------------------------------------------------------\
| Function:     GetCanMessage
|
| Purpose:      Get a message from the CAN RX queue.
|
| Discussion:   Disables CAN RX Buffer Full interrupts for <N> cycles.
|
| Expects:      Nothing.
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from serial code.
\---------------------------------------------------------------------------*/

void GetCanMessage (CanMessage_t* receivedMsg)
{
    static  bit     rx0IntsEnabled;
    static  bit     rx1IntsEnabled;
    static  uint8_t i;

    // Store the current queue index - don't need to turn off interrupts since
    // the CanRxIsr won't overwrite messages in the queue (except the very
    // last one.

    receivedMsg->header.bytes.low = CanRxQueue[canRxIndexOut].header.bytes.low;
    receivedMsg->header.bytes.high = CanRxQueue[canRxIndexOut].header.bytes.high;

    // Copy over all data bytes

    for (i = 0; i < CanRxQueue[canRxIndexOut].header.bits.length; ++i)
        receivedMsg->data[i] =  CanRxQueue[canRxIndexOut].data[i];

    //  Increment the index to the CIRCULAR queue.

    if (canRxIndexOut < (CAN_RX_QUEUE_SIZE-1))
        ++canRxIndexOut;
    else
        canRxIndexOut = 0;

    // We're about to check to see if there are any more message in the queue.
    // If there aren't any, we clear the "message available" flag. However,
    // it would really suck if we got a message in the queue *right* before
    // the clear instruction and then ignored the received message until we
    // got another message which sets the bit. Thus, we'll temporarily disable
    // the CAN receive interrupts - paranoid but necessary.

    if ( CanRxb1InterruptEnabled == TRUE)
    {
        rx1IntsEnabled = TRUE;
        DisableCanRx1Ints();
    }
    else
        rx1IntsEnabled = FALSE;


    if ( CanRxb0InterruptEnabled == TRUE)
    {
        rx0IntsEnabled = TRUE;
        DisableCanRx0Ints();
    }
    else
        rx0IntsEnabled = FALSE;

    // If the in and out indices are equal, there ain't no more messages so
    // clear the available flag.

    if (canRxIndexOut == canRxIndexIn)
        canMessageAvailable = 0;

    // Re-enable only the receive interrupts which were already enabled.
    // Don't make the mistake of ALWAYS enabling the interrupts.

    if (rx0IntsEnabled == TRUE)
        EnableCanRx0Ints();

    if (rx1IntsEnabled == TRUE)
        EnableCanRx1Ints();
}


/*---------------------------------------------------------------------------\
| Function:     CanRx0Isr
|
| Purpose:      Handle the CAN Receive interrupt for RXB0 (Receive Buffer 0)
|
| Discussion:   Receive stuff. Trow it in the CanRxQueue
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from serial code.
\---------------------------------------------------------------------------*/

#if INTS_CAN_RX0_ISR_USED == INTS_CORE_ISR_ENABLED

void CanRx0Isr (void)
{
    static  uint8_t*    pRegister; //static volatile near unsigned char RXB0D0      @ 0xF66;
    static  uint8_t     i;

    // Transfer the message from the CAN message buffer to the receive queue.
    //  Note that we're not going to bother to setup CANCON to move the RXB0
    // registers into the access bank - although for future renditions this
    // could be a good idea if it makes sense in C.

    // Copy over the message ID

    CanRxQueue[canRxIndexIn].header.bytes.high = RXB0SIDH;
    CanRxQueue[canRxIndexIn].header.bytes.low = RXB0SIDL & 0b11100000;

    // Copy over the DLC and RTR bits & check the DLC bits

    CanRxQueue[canRxIndexIn].header.bytes.low |= RXB0DLC & 0b00001111;
    CanRxQueue[canRxIndexIn].header.bits.request = RXB0RTRRO;

    if ( CanRxQueue[canRxIndexIn].header.bits.length > 8 )
    {
        #if INTS_CAN_RX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(CAN_RX0ISR_BAD_DLC);
        #else
        ThrowWarning_LP(CAN_RX0ISR_BAD_DLC);
        #endif

        CanRxQueue[canRxIndexIn].header.bits.length = 8;
    }

    // Copy over all data bytes

    pRegister = &RXB0D0;

    for (i = 0; i < CanRxQueue[canRxIndexIn].header.bits.length; ++i)
        CanRxQueue[canRxIndexIn].data[i] = *pRegister++;

    // We've got the message; clear interrupt (& enable a new message to be
    // loaded in the buffer while we do other things.)

    ClearCanRx0Interrupt();

    //  Increment the index to the CIRCULAR queue.

    if (canRxIndexIn < (CAN_RX_QUEUE_SIZE-1))
        ++canRxIndexIn;
    else
        canRxIndexIn = 0;

    // Check for overwrites. If we incremented the indexIn and it now
    // equals the indexOut then we've wrapped around the queue and
    // we have a queue overrun. In this case, don't write to the
    // queue - back off by one and throw a warning.

    if (canRxIndexIn == canRxIndexOut)
    {
        if (canRxIndexIn == 0)
            canRxIndexIn  = (CAN_RX_QUEUE_SIZE-1);
        else
            --canRxIndexIn;

        #if INTS_CAN_RX0_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(CAN_RX_QUEUE_OVERRUN_RXB0);
        #else
        ThrowWarning_LP(CAN_RX_QUEUE_OVERRUN_RXB0);
        #endif
    }

    // In any case, let everyone know that there's data to be had in the queue

    canMessageAvailable = 1;
}


#endif // INTS_CAN_RX0_ISR_USED == INTS_CORE_ISR_ENABLED


/*---------------------------------------------------------------------------\
| Function:     CanRx1Isr
|
| Purpose:      Handle the CAN Receive interrupt for RXB1 (Receive Buffer 1)
|
| Discussion:   Receive stuff. Trow it in the CanRxQueue
|
| Expects:      -
|
|    Date    | Notes
|------------|------------------------------------------------------------
| 06/01/2002 | Stolen from serial code.
\---------------------------------------------------------------------------*/

#if INTS_CAN_RX1_ISR_USED == INTS_CORE_ISR_ENABLED

void CanRx1Isr (void)
{

    static  uint8_t*    pRegister;
    static  uint8_t     i;

    // Transfer the message from the CAN message buffer to the receive queue.
    //  Note that we're not going to bother to setup CANCON to move the RXB1
    // registers into the access bank - although for future renditions this
    // could be a good idea if it makes sense in C.



    // Copy over the message ID

    CanRxQueue[canRxIndexIn].header.bytes.high = RXB1SIDH;
    CanRxQueue[canRxIndexIn].header.bytes.low = RXB1SIDL & 0b11100000;

    // Copy over the DLC and RTR bits & check the DLC bits

    CanRxQueue[canRxIndexIn].header.bytes.low |= RXB1DLC & 0b00001111;
    CanRxQueue[canRxIndexIn].header.bits.request = RXB1RTRRO;

    if ( CanRxQueue[canRxIndexIn].header.bits.length > 8 )
    {
        #if INTS_CAN_RX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(CAN_RX1ISR_BAD_DLC);
        #else
        ThrowWarning_LP(CAN_RX1ISR_BAD_DLC);
        #endif

        CanRxQueue[canRxIndexIn].header.bits.length = 8;
    }

    // Copy over all data bytes

    pRegister = &RXB1D0;

    for (i = 0; i < CanRxQueue[canRxIndexIn].header.bits.length; ++i)
        CanRxQueue[canRxIndexIn].data[i] = *pRegister++;

    // We've got the message; clear interrupt (& enable a new message to be
    // loaded in the buffer while we do other things.)

    ClearCanRx1Interrupt();

    //  Increment the index to the CIRCULAR queue.

    if (canRxIndexIn < (CAN_RX_QUEUE_SIZE-1))
        ++canRxIndexIn;
    else
        canRxIndexIn = 0;

    // Check for overwrites. If we incremented the indexIn and it now
    // equals the indexOut then we've wrapped around the queue and
    // we have a queue overrun. In this case, don't write to the
    // queue - back off by one and throw a warning.

    if (canRxIndexIn == canRxIndexOut)
    {
        if (canRxIndexIn == 0)
            canRxIndexIn  = (CAN_RX_QUEUE_SIZE-1);
        else
            --canRxIndexIn;

        #if INTS_CAN_RX1_INT_PRIORITY == INTS_HIGH_PRIORITY_INTERRUPT
        ThrowWarning_HP(CAN_RX_QUEUE_OVERRUN_RXB1);
        #else
        ThrowWarning_LP(CAN_RX_QUEUE_OVERRUN_RXB1);
        #endif
    }

    // In any case, let everyone know that there's data to be had in the queue

    canMessageAvailable = 1;
}

#endif //INTS_CAN_RX0_ISR_USED == INTS_CORE_ISR_ENABLED


